/**
 * @file qmi8658c.c
 * @brief Implementation file for QMI8658C IMU driver
 * @author Generated
 * @date 2023
 */

#include "qmi8658c.h"

/* STM32 HAL dependencies */
#include "stm32f4xx_hal.h"  // Modify to match your STM32 family

/* Static variables */
static I2C_HandleTypeDef *QMI8658C_hi2c;  // I2C handle pointer
static uint8_t QMI8658C_Address = QMI8658C_I2C_ADDR;
static QMI8658C_Config_t QMI8658C_CurrentConfig;
static uint16_t accelSensitivity = 16384;  // Default ±2g full scale (LSB/g)
static uint16_t gyroSensitivity = 2048;    // Default ±16dps full scale (LSB/dps)

/* Private function prototypes */
static QMI8658C_Status_t QMI8658C_ReadReg(uint8_t reg, uint8_t *data, uint16_t len);
static QMI8658C_Status_t QMI8658C_WriteReg(uint8_t reg, uint8_t *data, uint16_t len);
static QMI8658C_Status_t QMI8658C_UpdateSensitivities(void);

/* Implementation */

/**
 * @brief Initialize QMI8658C device with default configuration
 */
QMI8658C_Status_t QMI8658C_Init(void *hi2c)
{
    uint8_t id;
    
    /* Store I2C handle */
    QMI8658C_hi2c = (I2C_HandleTypeDef *)hi2c;
    
    /* Read WHO_AM_I register to verify device ID */
    if (QMI8658C_ReadID(&id) != QMI8658C_OK) {
        return QMI8658C_ERROR_COMM;
    }
    
    /* Check if device ID is correct */
    if (id != QMI8658C_WHO_AM_I) {
        return QMI8658C_ERROR_ID;
    }
    
    /* Reset device */
    if (QMI8658C_Reset() != QMI8658C_OK) {
        return QMI8658C_ERROR_INIT;
    }
    
    /* Wait for reset to complete */
    HAL_Delay(50);
    
    /* Set default configuration */
    return QMI8658C_SetDefaultConfig();
}

/**
 * @brief Initialize QMI8658C device with provided configuration
 */
QMI8658C_Status_t QMI8658C_InitWithConfig(void *hi2c, QMI8658C_Config_t *config)
{
    QMI8658C_Status_t status;
    
    /* First initialize with default settings */
    status = QMI8658C_Init(hi2c);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Store the current configuration */
    QMI8658C_CurrentConfig = *config;

    /* Apply configuration */
    
    /* Configure CTRL1 register */
    uint8_t ctrl1 = 0;
    if (config->syncSampleEnabled) {
        ctrl1 |= QMI8658C_CTRL1_SPI_AI;
    }
    ctrl1 |= QMI8658C_CTRL1_SPI_BE;  /* Use big endian by default */
    
    status = QMI8658C_WriteReg(QMI8658C_REG_CTRL1, &ctrl1, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Configure accelerometer if enabled */
    if (config->accelEnabled) {
        status = QMI8658C_EnableAccel(config->accelODR, config->accelRange);
        if (status != QMI8658C_OK) {
            return status;
        }
    }
    
    /* Configure gyroscope if enabled */
    if (config->gyroEnabled) {
        status = QMI8658C_EnableGyro(config->gyroODR, config->gyroRange);
        if (status != QMI8658C_OK) {
            return status;
        }
    }
    
    /* Configure magnetometer if enabled */
    if (config->magEnabled) {
        status = QMI8658C_EnableMag(config->magDevice, config->magODR);
        if (status != QMI8658C_OK) {
            return status;
        }
    }
    
    /* Configure filters */
    status = QMI8658C_ConfigAccelLPF(config->accelLPFEnabled, config->accelLPFMode);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    status = QMI8658C_ConfigGyroLPF(config->gyroLPFEnabled, config->gyroLPFMode);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Configure AttitudeEngine if enabled */
    if (config->aeEnabled) {
        status = QMI8658C_EnableAttitudeEngine(config->aeODR, config->aeMoDEnabled);
        if (status != QMI8658C_OK) {
            return status;
        }
    }
    
    /* Configure FIFO if not in bypass mode */
    if (config->fifoMode != QMI8658C_FIFO_MODE_BYPASS) {
        status = QMI8658C_ConfigFIFO(config->fifoMode, config->fifoSize, config->fifoWatermark);
        if (status != QMI8658C_OK) {
            return status;
        }
    }
    
    /* Configure CTRL7 register with sensor enables */
    uint8_t ctrl7 = 0;
    if (config->syncSampleEnabled) {
        ctrl7 |= QMI8658C_CTRL7_SYNC_SMPL;
    }
    if (config->highSpeedClock) {
        ctrl7 |= QMI8658C_CTRL7_SYS_HS;
    }
    if (config->gyroSnoozeEnabled) {
        ctrl7 |= QMI8658C_CTRL7_G_SNOOZE;
    }
    if (config->aeEnabled) {
        ctrl7 |= QMI8658C_CTRL7_AE_EN;
    }
    if (config->magEnabled) {
        ctrl7 |= QMI8658C_CTRL7_MAG_EN;
    }
    if (config->gyroEnabled) {
        ctrl7 |= QMI8658C_CTRL7_GYRO_EN;
    }
    if (config->accelEnabled) {
        ctrl7 |= QMI8658C_CTRL7_ACCEL_EN;
    }
    
    status = QMI8658C_WriteReg(QMI8658C_REG_CTRL7, &ctrl7, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Update sensitivities based on the current configuration */
    QMI8658C_UpdateSensitivities();
    
    return QMI8658C_OK;
}

/**
 * @brief Reset QMI8658C device
 */
QMI8658C_Status_t QMI8658C_Reset(void)
{
    uint8_t resetValue = 0x01;  // Reset value
    
    /* Perform soft reset */
    return QMI8658C_WriteReg(QMI8658C_REG_RESET, &resetValue, 1);
}

/**
 * @brief Set QMI8658C device to default configuration
 */
QMI8658C_Status_t QMI8658C_SetDefaultConfig(void)
{
    QMI8658C_Config_t defaultConfig;
    
    /* Default configuration values */
    defaultConfig.accelRange = QMI8658C_ACCEL_RANGE_2G;
    defaultConfig.accelODR = QMI8658C_ACCEL_ODR_125HZ;
    defaultConfig.gyroRange = QMI8658C_GYRO_RANGE_16DPS;
    defaultConfig.gyroODR = QMI8658C_GYRO_ODR_125HZ;
    defaultConfig.magDevice = QMI8658C_MAG_NONE;
    defaultConfig.magODR = QMI8658C_MAG_ODR_125HZ;
    defaultConfig.accelLPFMode = QMI8658C_LPF_MODE_0;
    defaultConfig.gyroLPFMode = QMI8658C_LPF_MODE_0;
    defaultConfig.accelLPFEnabled = false;
    defaultConfig.gyroLPFEnabled = false;
    defaultConfig.aeODR = QMI8658C_AE_ODR_16HZ;
    defaultConfig.aeMoDEnabled = false;
    defaultConfig.syncSampleEnabled = false;
    defaultConfig.highSpeedClock = false;
    defaultConfig.gyroSnoozeEnabled = false;
    defaultConfig.aeEnabled = false;
    defaultConfig.magEnabled = false;
    defaultConfig.gyroEnabled = true;
    defaultConfig.accelEnabled = true;
    defaultConfig.fifoSize = QMI8658C_FIFO_SIZE_16;
    defaultConfig.fifoMode = QMI8658C_FIFO_MODE_BYPASS;
    defaultConfig.fifoWatermark = 1;
    
    /* Apply default configuration */
    return QMI8658C_InitWithConfig(QMI8658C_hi2c, &defaultConfig);
}

/**
 * @brief Read QMI8658C device ID (WHO_AM_I register)
 */
QMI8658C_Status_t QMI8658C_ReadID(uint8_t *id)
{
    return QMI8658C_ReadReg(QMI8658C_REG_WHO_AM_I, id, 1);
}

/**
 * @brief Enable accelerometer
 */
QMI8658C_Status_t QMI8658C_EnableAccel(QMI8658C_AccelODR_t odr, QMI8658C_AccelRange_t range)
{
    uint8_t ctrl2;
    
    /* Construct CTRL2 register value */
    ctrl2 = ((uint8_t)range << 4) | (uint8_t)odr;
    
    /* Write CTRL2 register */
    QMI8658C_Status_t status = QMI8658C_WriteReg(QMI8658C_REG_CTRL2, &ctrl2, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Update current configuration */
    QMI8658C_CurrentConfig.accelODR = odr;
    QMI8658C_CurrentConfig.accelRange = range;
    QMI8658C_CurrentConfig.accelEnabled = true;
    
    /* Update sensitivities */
    QMI8658C_UpdateSensitivities();
    
    /* Update CTRL7 to enable accelerometer */
    uint8_t ctrl7;
    status = QMI8658C_ReadReg(QMI8658C_REG_CTRL7, &ctrl7, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    ctrl7 |= QMI8658C_CTRL7_ACCEL_EN;
    
    return QMI8658C_WriteReg(QMI8658C_REG_CTRL7, &ctrl7, 1);
}

/**
 * @brief Disable accelerometer
 */
QMI8658C_Status_t QMI8658C_DisableAccel(void)
{
    uint8_t ctrl7;
    
    /* Read current CTRL7 register */
    QMI8658C_Status_t status = QMI8658C_ReadReg(QMI8658C_REG_CTRL7, &ctrl7, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Clear accelerometer enable bit */
    ctrl7 &= ~QMI8658C_CTRL7_ACCEL_EN;
    
    /* Write updated CTRL7 register */
    status = QMI8658C_WriteReg(QMI8658C_REG_CTRL7, &ctrl7, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Update current configuration */
    QMI8658C_CurrentConfig.accelEnabled = false;
    
    return QMI8658C_OK;
}

/**
 * @brief Enable gyroscope
 */
QMI8658C_Status_t QMI8658C_EnableGyro(QMI8658C_GyroODR_t odr, QMI8658C_GyroRange_t range)
{
    uint8_t ctrl3;
    
    /* Construct CTRL3 register value */
    ctrl3 = ((uint8_t)range << 4) | (uint8_t)odr;
    
    /* Write CTRL3 register */
    QMI8658C_Status_t status = QMI8658C_WriteReg(QMI8658C_REG_CTRL3, &ctrl3, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Update current configuration */
    QMI8658C_CurrentConfig.gyroODR = odr;
    QMI8658C_CurrentConfig.gyroRange = range;
    QMI8658C_CurrentConfig.gyroEnabled = true;
    
    /* Update sensitivities */
    QMI8658C_UpdateSensitivities();
    
    /* Update CTRL7 to enable gyroscope */
    uint8_t ctrl7;
    status = QMI8658C_ReadReg(QMI8658C_REG_CTRL7, &ctrl7, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    ctrl7 |= QMI8658C_CTRL7_GYRO_EN;
    
    return QMI8658C_WriteReg(QMI8658C_REG_CTRL7, &ctrl7, 1);
}

/**
 * @brief Disable gyroscope
 */
QMI8658C_Status_t QMI8658C_DisableGyro(void)
{
    uint8_t ctrl7;
    
    /* Read current CTRL7 register */
    QMI8658C_Status_t status = QMI8658C_ReadReg(QMI8658C_REG_CTRL7, &ctrl7, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Clear gyroscope enable bit */
    ctrl7 &= ~QMI8658C_CTRL7_GYRO_EN;
    
    /* Write updated CTRL7 register */
    status = QMI8658C_WriteReg(QMI8658C_REG_CTRL7, &ctrl7, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Update current configuration */
    QMI8658C_CurrentConfig.gyroEnabled = false;
    
    return QMI8658C_OK;
}

/**
 * @brief Enable magnetometer
 */
QMI8658C_Status_t QMI8658C_EnableMag(QMI8658C_MagDevice_t device, QMI8658C_MagODR_t odr)
{
    uint8_t ctrl4;
    
    /* Construct CTRL4 register value */
    ctrl4 = ((uint8_t)device << 3) | (uint8_t)odr;
    
    /* Write CTRL4 register */
    QMI8658C_Status_t status = QMI8658C_WriteReg(QMI8658C_REG_CTRL4, &ctrl4, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Update current configuration */
    QMI8658C_CurrentConfig.magDevice = device;
    QMI8658C_CurrentConfig.magODR = odr;
    QMI8658C_CurrentConfig.magEnabled = true;
    
    /* Update CTRL7 to enable magnetometer */
    uint8_t ctrl7;
    status = QMI8658C_ReadReg(QMI8658C_REG_CTRL7, &ctrl7, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    ctrl7 |= QMI8658C_CTRL7_MAG_EN;
    
    return QMI8658C_WriteReg(QMI8658C_REG_CTRL7, &ctrl7, 1);
}

/**
 * @brief Disable magnetometer
 */
QMI8658C_Status_t QMI8658C_DisableMag(void)
{
    uint8_t ctrl7;
    
    /* Read current CTRL7 register */
    QMI8658C_Status_t status = QMI8658C_ReadReg(QMI8658C_REG_CTRL7, &ctrl7, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Clear magnetometer enable bit */
    ctrl7 &= ~QMI8658C_CTRL7_MAG_EN;
    
    /* Write updated CTRL7 register */
    status = QMI8658C_WriteReg(QMI8658C_REG_CTRL7, &ctrl7, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Update current configuration */
    QMI8658C_CurrentConfig.magEnabled = false;
    
    return QMI8658C_OK;
}

/**
 * @brief Enable AttitudeEngine
 */
QMI8658C_Status_t QMI8658C_EnableAttitudeEngine(QMI8658C_AE_ODR_t odr, bool enableMoD)
{
    uint8_t ctrl6;
    
    /* Construct CTRL6 register value */
    ctrl6 = (enableMoD ? 0x80 : 0x00) | (uint8_t)odr;
    
    /* Write CTRL6 register */
    QMI8658C_Status_t status = QMI8658C_WriteReg(QMI8658C_REG_CTRL6, &ctrl6, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Update current configuration */
    QMI8658C_CurrentConfig.aeODR = odr;
    QMI8658C_CurrentConfig.aeMoDEnabled = enableMoD;
    QMI8658C_CurrentConfig.aeEnabled = true;
    
    /* Update CTRL7 to enable AttitudeEngine */
    uint8_t ctrl7;
    status = QMI8658C_ReadReg(QMI8658C_REG_CTRL7, &ctrl7, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    ctrl7 |= QMI8658C_CTRL7_AE_EN;
    
    /* Ensure gyro and accel are also enabled */
    ctrl7 |= (QMI8658C_CTRL7_GYRO_EN | QMI8658C_CTRL7_ACCEL_EN);
    
    return QMI8658C_WriteReg(QMI8658C_REG_CTRL7, &ctrl7, 1);
}

/**
 * @brief Disable AttitudeEngine
 */
QMI8658C_Status_t QMI8658C_DisableAttitudeEngine(void)
{
    uint8_t ctrl7;
    
    /* Read current CTRL7 register */
    QMI8658C_Status_t status = QMI8658C_ReadReg(QMI8658C_REG_CTRL7, &ctrl7, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Clear AttitudeEngine enable bit */
    ctrl7 &= ~QMI8658C_CTRL7_AE_EN;
    
    /* Write updated CTRL7 register */
    status = QMI8658C_WriteReg(QMI8658C_REG_CTRL7, &ctrl7, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Update current configuration */
    QMI8658C_CurrentConfig.aeEnabled = false;
    
    return QMI8658C_OK;
}

/**
 * @brief Configure accelerometer low-pass filter
 */
QMI8658C_Status_t QMI8658C_ConfigAccelLPF(bool enable, QMI8658C_LPFMode_t mode)
{
    uint8_t ctrl5;
    
    /* Read current CTRL5 register */
    QMI8658C_Status_t status = QMI8658C_ReadReg(QMI8658C_REG_CTRL5, &ctrl5, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Update CTRL5 register */
    ctrl5 &= ~0x07;  // Clear aLPF_EN and aLPF_MODE bits
    ctrl5 |= ((uint8_t)mode << 1);
    if (enable) {
        ctrl5 |= 0x01;  // Set aLPF_EN bit
    }
    
    /* Write updated CTRL5 register */
    status = QMI8658C_WriteReg(QMI8658C_REG_CTRL5, &ctrl5, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Update current configuration */
    QMI8658C_CurrentConfig.accelLPFEnabled = enable;
    QMI8658C_CurrentConfig.accelLPFMode = mode;
    
    return QMI8658C_OK;
}

/**
 * @brief Configure gyroscope low-pass filter
 */
QMI8658C_Status_t QMI8658C_ConfigGyroLPF(bool enable, QMI8658C_LPFMode_t mode)
{
    uint8_t ctrl5;
    
    /* Read current CTRL5 register */
    QMI8658C_Status_t status = QMI8658C_ReadReg(QMI8658C_REG_CTRL5, &ctrl5, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Update CTRL5 register */
    ctrl5 &= ~0x70;  // Clear gLPF_EN and gLPF_MODE bits
    ctrl5 |= ((uint8_t)mode << 5);
    if (enable) {
        ctrl5 |= 0x10;  // Set gLPF_EN bit
    }
    
    /* Write updated CTRL5 register */
    status = QMI8658C_WriteReg(QMI8658C_REG_CTRL5, &ctrl5, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Update current configuration */
    QMI8658C_CurrentConfig.gyroLPFEnabled = enable;
    QMI8658C_CurrentConfig.gyroLPFMode = mode;
    
    return QMI8658C_OK;
}

/**
 * @brief Configure FIFO
 */
QMI8658C_Status_t QMI8658C_ConfigFIFO(QMI8658C_FifoMode_t mode, QMI8658C_FifoSize_t size, uint8_t watermark)
{
    uint8_t fifoCtrl;
    
    /* Configure FIFO watermark threshold */
    QMI8658C_Status_t status = QMI8658C_WriteReg(QMI8658C_REG_FIFO_WTM_TH, &watermark, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Configure FIFO control register */
    fifoCtrl = ((uint8_t)size << 2) | (uint8_t)mode;
    
    /* Write FIFO_CTRL register */
    status = QMI8658C_WriteReg(QMI8658C_REG_FIFO_CTRL, &fifoCtrl, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Update current configuration */
    QMI8658C_CurrentConfig.fifoMode = mode;
    QMI8658C_CurrentConfig.fifoSize = size;
    QMI8658C_CurrentConfig.fifoWatermark = watermark;
    
    return QMI8658C_OK;
}

/**
 * @brief Reset FIFO
 */
QMI8658C_Status_t QMI8658C_ResetFIFO(void)
{
    return QMI8658C_ExecuteCommand(QMI8658C_CTRL_CMD_RST_FIFO);
}

/**
 * @brief Read data from FIFO
 */
QMI8658C_Status_t QMI8658C_ReadFIFO(uint8_t *buffer, uint16_t *count)
{
    uint8_t fifoStatus, fifoSmplCnt, fifoCtrl;
    uint16_t sampleCount;
    QMI8658C_Status_t status;
    
    /* Request FIFO access */
    status = QMI8658C_ExecuteCommand(QMI8658C_CTRL_CMD_REQ_FIFO);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Read FIFO_STATUS register */
    status = QMI8658C_ReadReg(QMI8658C_REG_FIFO_STATUS, &fifoStatus, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Read FIFO_SMPL_CNT register */
    status = QMI8658C_ReadReg(QMI8658C_REG_FIFO_SMPL_CNT, &fifoSmplCnt, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Calculate total sample count in bytes */
    sampleCount = ((uint16_t)(fifoStatus & 0x03) << 8) | fifoSmplCnt;
    
    /* Limit reading to buffer size */
    if (sampleCount > *count) {
        sampleCount = *count;
    }
    
    /* Read FIFO data */
    for (uint16_t i = 0; i < sampleCount; i++) {
        status = QMI8658C_ReadReg(QMI8658C_REG_FIFO_DATA, &buffer[i], 1);
        if (status != QMI8658C_OK) {
            *count = i;
            break;
        }
    }
    
    /* Clear FIFO read mode */
    status = QMI8658C_ReadReg(QMI8658C_REG_FIFO_CTRL, &fifoCtrl, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    fifoCtrl &= ~QMI8658C_FIFO_RD_MODE;
    
    status = QMI8658C_WriteReg(QMI8658C_REG_FIFO_CTRL, &fifoCtrl, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Return number of bytes read */
    *count = sampleCount;
    
    return QMI8658C_OK;
}

/**
 * @brief Read 6-axis IMU data (accelerometer + gyroscope)
 */
QMI8658C_Status_t QMI8658C_ReadIMUData(QMI8658C_IMUData_t *data)
{
    uint8_t buffer[14];  // Temp(2) + Accel(6) + Gyro(6) = 14 bytes
    
    /* Read all data at once to ensure coherent dataset */
    QMI8658C_Status_t status = QMI8658C_ReadReg(QMI8658C_REG_TEMP_L, buffer, 14);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Extract temperature (16-bit, 2's complement) */
    data->temp = (int16_t)((buffer[1] << 8) | buffer[0]);
    
    /* Extract accelerometer data (16-bit, 2's complement) */
    data->ax = (int16_t)((buffer[3] << 8) | buffer[2]);
    data->ay = (int16_t)((buffer[5] << 8) | buffer[4]);
    data->az = (int16_t)((buffer[7] << 8) | buffer[6]);
    
    /* Extract gyroscope data (16-bit, 2's complement) */
    data->gx = (int16_t)((buffer[9] << 8) | buffer[8]);
    data->gy = (int16_t)((buffer[11] << 8) | buffer[10]);
    data->gz = (int16_t)((buffer[13] << 8) | buffer[12]);
    
    return QMI8658C_OK;
}

/**
 * @brief Read 9-axis data (accelerometer + gyroscope + magnetometer)
 */
QMI8658C_Status_t QMI8658C_Read9DOFData(QMI8658C_9DOFData_t *data)
{
    uint8_t buffer[20];  // Temp(2) + Accel(6) + Gyro(6) + Mag(6) = 20 bytes
    
    /* Read all data at once to ensure coherent dataset */
    QMI8658C_Status_t status = QMI8658C_ReadReg(QMI8658C_REG_TEMP_L, buffer, 20);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Extract temperature (16-bit, 2's complement) */
    data->temp = (int16_t)((buffer[1] << 8) | buffer[0]);
    
    /* Extract accelerometer data (16-bit, 2's complement) */
    data->ax = (int16_t)((buffer[3] << 8) | buffer[2]);
    data->ay = (int16_t)((buffer[5] << 8) | buffer[4]);
    data->az = (int16_t)((buffer[7] << 8) | buffer[6]);
    
    /* Extract gyroscope data (16-bit, 2's complement) */
    data->gx = (int16_t)((buffer[9] << 8) | buffer[8]);
    data->gy = (int16_t)((buffer[11] << 8) | buffer[10]);
    data->gz = (int16_t)((buffer[13] << 8) | buffer[12]);
    
    /* Extract magnetometer data (16-bit, 2's complement) */
    data->mx = (int16_t)((buffer[15] << 8) | buffer[14]);
    data->my = (int16_t)((buffer[17] << 8) | buffer[16]);
    data->mz = (int16_t)((buffer[19] << 8) | buffer[18]);
    
    return QMI8658C_OK;
}

/**
 * @brief Read AttitudeEngine data (quaternion and velocity increments)
 */
QMI8658C_Status_t QMI8658C_ReadAttitudeEngineData(QMI8658C_AttitudeEngine_t *data)
{
    uint8_t buffer[16];  // dQW/X/Y/Z(8) + dVX/Y/Z(6) = 14 bytes
    
    /* If Motion on Demand mode is enabled, request data */
    if (QMI8658C_CurrentConfig.aeMoDEnabled) {
        QMI8658C_Status_t status = QMI8658C_ExecuteCommand(QMI8658C_CTRL_CMD_REQ_SDI);
        if (status != QMI8658C_OK) {
            return status;
        }
    }
    
    /* Read quaternion increments */
    QMI8658C_Status_t status = QMI8658C_ReadReg(QMI8658C_REG_DQW_L, buffer, 8);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Read velocity increments */
    status = QMI8658C_ReadReg(QMI8658C_REG_DVX_L, &buffer[8], 6);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Extract quaternion increments (16-bit, 2's complement) */
    data->dQW = (int16_t)((buffer[1] << 8) | buffer[0]);
    data->dQX = (int16_t)((buffer[3] << 8) | buffer[2]);
    data->dQY = (int16_t)((buffer[5] << 8) | buffer[4]);
    data->dQZ = (int16_t)((buffer[7] << 8) | buffer[6]);
    
    /* Extract velocity increments (16-bit, 2's complement) */
    data->dVX = (int16_t)((buffer[9] << 8) | buffer[8]);
    data->dVY = (int16_t)((buffer[11] << 8) | buffer[10]);
    data->dVZ = (int16_t)((buffer[13] << 8) | buffer[12]);
    
    return QMI8658C_OK;
}

/**
 * @brief Configure Wake on Motion (WoM) feature
 */
QMI8658C_Status_t QMI8658C_ConfigWakeOnMotion(uint8_t threshold, bool useINT2, 
                                              bool initialValue, uint8_t blankingTime)
{
    uint8_t cal1_l, cal1_h;
    QMI8658C_Status_t status;
    
    /* Disable sensors */
    uint8_t ctrl7 = 0x00;
    status = QMI8658C_WriteReg(QMI8658C_REG_CTRL7, &ctrl7, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Configure accelerometer in low power mode */
    uint8_t ctrl2 = (QMI8658C_ACCEL_RANGE_2G << 4) | QMI8658C_ACCEL_ODR_21HZ_LP;
    status = QMI8658C_WriteReg(QMI8658C_REG_CTRL2, &ctrl2, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Prepare CAL1_L register (threshold value in mg) */
    cal1_l = threshold;
    
    /* Prepare CAL1_H register (INT selection and blanking time) */
    cal1_h = blankingTime & 0x3F;  /* Bits 0-5: blanking time */
    
    if (useINT2) {
        cal1_h |= initialValue ? 0xC0 : 0x40;  /* Bits 6-7: 01=INT2(init=0), 11=INT2(init=1) */
    } else {
        cal1_h |= initialValue ? 0x80 : 0x00;  /* Bits 6-7: 00=INT1(init=0), 10=INT1(init=1) */
    }
    
    /* Write CAL1_L and CAL1_H registers */
    status = QMI8658C_WriteReg(QMI8658C_REG_CAL1_L, &cal1_l, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    status = QMI8658C_WriteReg(QMI8658C_REG_CAL1_H, &cal1_h, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Execute WoM setting command */
    status = QMI8658C_ExecuteCommand(QMI8658C_CTRL_CMD_WRITE_WOM_SETTING);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Enable accelerometer */
    ctrl7 = QMI8658C_CTRL7_ACCEL_EN;
    status = QMI8658C_WriteReg(QMI8658C_REG_CTRL7, &ctrl7, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    return QMI8658C_OK;
}

/**
 * @brief Disable Wake on Motion feature
 */
QMI8658C_Status_t QMI8658C_DisableWakeOnMotion(void)
{
    uint8_t cal1_l = 0x00;  /* Threshold value of 0 disables WoM */
    QMI8658C_Status_t status;
    
    /* Disable sensors */
    uint8_t ctrl7 = 0x00;
    status = QMI8658C_WriteReg(QMI8658C_REG_CTRL7, &ctrl7, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Write CAL1_L register */
    status = QMI8658C_WriteReg(QMI8658C_REG_CAL1_L, &cal1_l, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Execute WoM setting command */
    return QMI8658C_ExecuteCommand(QMI8658C_CTRL_CMD_WRITE_WOM_SETTING);
}

/**
 * @brief Perform accelerometer self-test
 */
QMI8658C_Status_t QMI8658C_AccelSelfTest(bool *result)
{
    QMI8658C_IMUData_t normalData, testData;
    int16_t diffX, diffY, diffZ;
    uint8_t ctrl2;
    QMI8658C_Status_t status;
    
    /* Store current configuration */
    status = QMI8658C_ReadReg(QMI8658C_REG_CTRL2, &ctrl2, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Configure accelerometer for self-test */
    uint8_t accelConfig = (QMI8658C_ACCEL_RANGE_8G << 4) | QMI8658C_ACCEL_ODR_1000HZ;
    status = QMI8658C_WriteReg(QMI8658C_REG_CTRL2, &accelConfig, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Enable accelerometer */
    uint8_t ctrl7 = QMI8658C_CTRL7_ACCEL_EN;
    status = QMI8658C_WriteReg(QMI8658C_REG_CTRL7, &ctrl7, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Wait for stable output */
    HAL_Delay(100);
    
    /* Read normal mode data */
    status = QMI8658C_ReadIMUData(&normalData);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Enable self-test mode */
    accelConfig |= 0x80;  /* Set ST bit */
    status = QMI8658C_WriteReg(QMI8658C_REG_CTRL2, &accelConfig, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Wait for self-test effect to appear */
    HAL_Delay(100);
    
    /* Read self-test mode data */
    status = QMI8658C_ReadIMUData(&testData);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Restore original configuration */
    status = QMI8658C_WriteReg(QMI8658C_REG_CTRL2, &ctrl2, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Calculate differences */
    diffX = testData.ax - normalData.ax;
    diffY = testData.ay - normalData.ay;
    diffZ = testData.az - normalData.az;
    
    /* Self-test criteria (simplified - manufacturer should provide exact thresholds) */
    *result = ((diffX > 1000) && (diffY > 1000) && (diffZ > 1000));
    
    return QMI8658C_OK;
}

/**
 * @brief Perform gyroscope self-test
 */
QMI8658C_Status_t QMI8658C_GyroSelfTest(bool *result)
{
    QMI8658C_IMUData_t normalData, testData;
    int16_t diffX, diffY, diffZ;
    uint8_t ctrl3;
    QMI8658C_Status_t status;
    
    /* Store current configuration */
    status = QMI8658C_ReadReg(QMI8658C_REG_CTRL3, &ctrl3, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Configure gyroscope for self-test */
    uint8_t gyroConfig = (QMI8658C_GYRO_RANGE_256DPS << 4) | QMI8658C_GYRO_ODR_1000HZ;
    status = QMI8658C_WriteReg(QMI8658C_REG_CTRL3, &gyroConfig, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Enable gyroscope */
    uint8_t ctrl7 = QMI8658C_CTRL7_GYRO_EN;
    status = QMI8658C_WriteReg(QMI8658C_REG_CTRL7, &ctrl7, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Wait for stable output */
    HAL_Delay(100);
    
    /* Read normal mode data */
    status = QMI8658C_ReadIMUData(&normalData);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Enable self-test mode */
    gyroConfig |= 0x80;  /* Set ST bit */
    status = QMI8658C_WriteReg(QMI8658C_REG_CTRL3, &gyroConfig, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Wait for self-test effect to appear */
    HAL_Delay(100);
    
    /* Read self-test mode data */
    status = QMI8658C_ReadIMUData(&testData);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Restore original configuration */
    status = QMI8658C_WriteReg(QMI8658C_REG_CTRL3, &ctrl3, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Calculate differences */
    diffX = testData.gx - normalData.gx;
    diffY = testData.gy - normalData.gy;
    diffZ = testData.gz - normalData.gz;
    
    /* Self-test criteria (simplified - manufacturer should provide exact thresholds) */
    *result = ((diffX > 1000) && (diffY > 1000) && (diffZ > 1000));
    
    return QMI8658C_OK;
}

/**
 * @brief Set gyroscope bias
 */
QMI8658C_Status_t QMI8658C_SetGyroBias(int16_t xBias, int16_t yBias, int16_t zBias)
{
    uint8_t biasBuffer[6];
    QMI8658C_Status_t status;
    
    /* Prepare bias values in correct format */
    biasBuffer[0] = (uint8_t)(xBias & 0xFF);
    biasBuffer[1] = (uint8_t)((xBias >> 8) & 0xFF);
    biasBuffer[2] = (uint8_t)(yBias & 0xFF);
    biasBuffer[3] = (uint8_t)((yBias >> 8) & 0xFF);
    biasBuffer[4] = (uint8_t)(zBias & 0xFF);
    biasBuffer[5] = (uint8_t)((zBias >> 8) & 0xFF);
    
    /* Write bias values to CAL registers */
    status = QMI8658C_WriteReg(QMI8658C_REG_CAL1_L, &biasBuffer[0], 2);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    status = QMI8658C_WriteReg(QMI8658C_REG_CAL2_L, &biasBuffer[2], 2);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    status = QMI8658C_WriteReg(QMI8658C_REG_CAL3_L, &biasBuffer[4], 2);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Execute gyro bias command */
    return QMI8658C_ExecuteCommand(QMI8658C_CTRL_CMD_GYRO_BIAS);
}

/**
 * @brief Set accelerometer bias
 */
QMI8658C_Status_t QMI8658C_SetAccelBias(int16_t xBias, int16_t yBias, int16_t zBias)
{
    uint8_t biasBuffer[6];
    QMI8658C_Status_t status;
    
    /* Prepare bias values in correct format */
    biasBuffer[0] = (uint8_t)(xBias & 0xFF);
    biasBuffer[1] = (uint8_t)((xBias >> 8) & 0xFF);
    biasBuffer[2] = (uint8_t)(yBias & 0xFF);
    biasBuffer[3] = (uint8_t)((yBias >> 8) & 0xFF);
    biasBuffer[4] = (uint8_t)(zBias & 0xFF);
    biasBuffer[5] = (uint8_t)((zBias >> 8) & 0xFF);
    
    /* Write bias values to CAL registers */
    status = QMI8658C_WriteReg(QMI8658C_REG_CAL1_L, &biasBuffer[0], 2);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    status = QMI8658C_WriteReg(QMI8658C_REG_CAL2_L, &biasBuffer[2], 2);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    status = QMI8658C_WriteReg(QMI8658C_REG_CAL3_L, &biasBuffer[4], 2);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Execute accel bias command */
    return QMI8658C_ExecuteCommand(QMI8658C_CTRL_CMD_ACCEL_HOST_DELTA_OFFSET);
}

/**
 * @brief Enter power-down mode
 */
QMI8658C_Status_t QMI8658C_EnterPowerDown(void)
{
    uint8_t ctrl1;
    
    /* Read current CTRL1 register */
    QMI8658C_Status_t status = QMI8658C_ReadReg(QMI8658C_REG_CTRL1, &ctrl1, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Set sensor disable bit */
    ctrl1 |= QMI8658C_CTRL1_SENSOR_DISABLE;
    
    /* Write updated CTRL1 register */
    return QMI8658C_WriteReg(QMI8658C_REG_CTRL1, &ctrl1, 1);
}

/**
 * @brief Exit power-down mode
 */
QMI8658C_Status_t QMI8658C_ExitPowerDown(void)
{
    uint8_t ctrl1;
    
    /* Read current CTRL1 register */
    QMI8658C_Status_t status = QMI8658C_ReadReg(QMI8658C_REG_CTRL1, &ctrl1, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Clear sensor disable bit */
    ctrl1 &= ~QMI8658C_CTRL1_SENSOR_DISABLE;
    
    /* Write updated CTRL1 register */
    return QMI8658C_WriteReg(QMI8658C_REG_CTRL1, &ctrl1, 1);
}

/**
 * @brief Execute CTRL9 command and wait for completion
 */
QMI8658C_Status_t QMI8658C_ExecuteCommand(uint8_t command)
{
    uint8_t status1;
    QMI8658C_Status_t status;
    
    /* Write command to CTRL9 register */
    status = QMI8658C_WriteReg(QMI8658C_REG_CTRL9, &command, 1);
    if (status != QMI8658C_OK) {
        return status;
    }
    
    /* Wait for command to complete (with timeout) */
    uint16_t timeout = 1000;  // 1 second timeout
    do {
        HAL_Delay(1);
        status = QMI8658C_ReadReg(QMI8658C_REG_STATUS1, &status1, 1);
        if (status != QMI8658C_OK) {
            return status;
        }
        
        timeout--;
        if (timeout == 0) {
            return QMI8658C_ERROR;
        }
    } while ((status1 & 0x01) == 0);  // Wait for CmdDone bit
    
    return QMI8658C_OK;
}

/* Private function implementations */

/**
 * @brief Read register from QMI8658C
 */
static QMI8658C_Status_t QMI8658C_ReadReg(uint8_t reg, uint8_t *data, uint16_t len)
{
    HAL_StatusTypeDef hal_status;
    
    hal_status = HAL_I2C_Mem_Read(QMI8658C_hi2c, QMI8658C_Address << 1, reg, 
                                   I2C_MEMADD_SIZE_8BIT, data, len, 1000);
    
    return (hal_status == HAL_OK) ? QMI8658C_OK : QMI8658C_ERROR_COMM;
}

/**
 * @brief Write register to QMI8658C
 */
static QMI8658C_Status_t QMI8658C_WriteReg(uint8_t reg, uint8_t *data, uint16_t len)
{
    HAL_StatusTypeDef hal_status;
    
    hal_status = HAL_I2C_Mem_Write(QMI8658C_hi2c, QMI8658C_Address << 1, reg, 
                                    I2C_MEMADD_SIZE_8BIT, data, len, 1000);
    
    return (hal_status == HAL_OK) ? QMI8658C_OK : QMI8658C_ERROR_COMM;
}

/**
 * @brief Update sensitivity values based on current configuration
 */
static QMI8658C_Status_t QMI8658C_UpdateSensitivities(void)
{
    /* Update accelerometer sensitivity based on range */
    switch (QMI8658C_CurrentConfig.accelRange) {
        case QMI8658C_ACCEL_RANGE_2G:
            accelSensitivity = 16384;  // LSB/g
            break;
        case QMI8658C_ACCEL_RANGE_4G:
            accelSensitivity = 8192;   // LSB/g
            break;
        case QMI8658C_ACCEL_RANGE_8G:
            accelSensitivity = 4096;   // LSB/g
            break;
        case QMI8658C_ACCEL_RANGE_16G:
            accelSensitivity = 2048;   // LSB/g
            break;
        default:
            return QMI8658C_ERROR;
    }
    
    /* Update gyroscope sensitivity based on range */
    switch (QMI8658C_CurrentConfig.gyroRange) {
        case QMI8658C_GYRO_RANGE_16DPS:
            gyroSensitivity = 2048;    // LSB/dps
            break;
        case QMI8658C_GYRO_RANGE_32DPS:
            gyroSensitivity = 1024;    // LSB/dps
            break;
        case QMI8658C_GYRO_RANGE_64DPS:
            gyroSensitivity = 512;     // LSB/dps
            break;
        case QMI8658C_GYRO_RANGE_128DPS:
            gyroSensitivity = 256;     // LSB/dps
            break;
        case QMI8658C_GYRO_RANGE_256DPS:
            gyroSensitivity = 128;     // LSB/dps
            break;
        case QMI8658C_GYRO_RANGE_512DPS:
            gyroSensitivity = 64;      // LSB/dps
            break;
        case QMI8658C_GYRO_RANGE_1024DPS:
            gyroSensitivity = 32;      // LSB/dps
            break;
        case QMI8658C_GYRO_RANGE_2048DPS:
            gyroSensitivity = 16;      // LSB/dps
            break;
        default:
            return QMI8658C_ERROR;
    }
    
    return QMI8658C_OK;
}
